#include <J3ML/Geometry/LineSegment.hpp>
#include <J3ML/Geometry/Sphere.hpp>
#include <J3ML/Geometry/TriangleMesh.hpp>

namespace J3ML::Geometry
{

    bool Sphere::Contains(const LineSegment &lineseg) const {
        return Contains(lineseg.A) && Contains(lineseg.B);
    }

    TriangleMesh Sphere::GenerateUVSphere(int subdivisions) const
    {
        // http://www.songho.ca/opengl/gl_sphere.html

        TriangleMesh mesh;

        float x, y, z, xy;                           // Vertex Position
        float nx, ny, nz, lengthInv = 1.f / Radius;  // Vertex Normal
        float s, t;                                  // Vertex TexCoord

        int sectorCount = subdivisions;
        int stackCount = subdivisions;

        float sectorStep = 2.f * Math::Pi / sectorCount;
        float stackStep = Math::Pi / stackCount;
        float sectorAngle, stackAngle;

        for (int i = 0; i <= stackCount; ++i)
        {
            stackAngle = Math::Pi / 2.f - i * stackStep;   // starting from pi/2 to -pi/2
            xy = Radius * Math::Cos(stackAngle);           // r * cos(u)
            z = Radius * Math::Sin(stackAngle);            // r * sin(u)

            // add (sectorCount + 1) vertices per stack
            // first and last vertices have same position and normal, but different tex coords
            for (int j = 0; j <= sectorCount; ++j)
            {
                sectorAngle = j * sectorStep;   // starting from 0 to 2pi

                // vertex position (x, y, z)
                x = xy * Math::Cos(sectorAngle);
                y = xy * Math::Sin(sectorAngle);

                Vector3 vertex = {x, y, z};

                mesh.Vertices.push_back(vertex);

                // normalized vertex normal (nx, ny, nz)
                nx = x * lengthInv;
                ny = y * lengthInv;
                nz = z * lengthInv;

                Vector3 normal = {nx, ny, nz};

                mesh.Normals.push_back(normal);

                // vertex tex coord (s, t) range between [0, 1]
                s = (float)j / sectorCount;
                t = (float)i / stackCount;

                Vector2 TexCoords = {s, t};

                mesh.UVs.push_back(normal);
            }
        }
        return mesh;
    }

    TriangleMesh Sphere::GenerateIcososphere() const
    {

        // Generate 12 vertices of an icosahedron for a given radius.

        const float h_angle = Math::Pi / 180.f * 72.f; // 72 degree = 360 / 5;
        const float v_angle = Math::Atan(1.f / 2.f);

        TriangleMesh mesh;

        int i1, i2;
        float z, xy;
        float hAngle1 = -Math::Pi / 2.f - h_angle / 2.f;
        float hAngle2 = -Math::Pi / 2;

        // the first top vertex at (0,0,r)
        Vector3 top_vertex = {0, 0, Radius};

        // compute 10 vertices at 1st and 2nd rows
        for (int i = 1; i <= 5; ++i)
        {
            i1 = i * 3;       // index for 1st row
            i2 = (i + 5) * 3; // index for 2nd row

            z  = Radius * Math::Sin(v_angle);   // elevation
            xy = Radius * Math::Cos(v_angle);   // length on XY plane

            Vector3 vert_0 = {xy * Math::Cos(hAngle1), xy * Math::Sin(hAngle1), z};
            Vector3 vert_1 = {xy * Math::Cos(hAngle2), xy * Math::Sin(hAngle2), -z};

            // next horizontal angles
            hAngle1 += h_angle;
            hAngle2 += h_angle;
        }

        // the last bottom vertex at (0, 0, -r)
        i1 = 11 * 3;
        Vector3 last_vertex = {0,0, -Radius};

        return mesh;
    }


    void Sphere::ProjectToAxis(const Vector3 &direction, float &outMin, float &outMax) const
    {
        float d = Vector3::Dot(direction, Position);
        outMin = d - Radius;
        outMax = d + Radius;
    }

    Vector3 Sphere::ExtremePoint(const Vector3 &direction) const {
        float len = direction.Length();
        assert(len > 0.f);
        return Position + direction * (Radius / len);
    }

    Vector3 Sphere::ExtremePoint(const Vector3 &direction, float &projectionDistance) const {
        Vector3 extremePoint = ExtremePoint(direction);
        projectionDistance = extremePoint.Dot(direction);
        return extremePoint;
    }

    void Sphere::Translate(const Vector3 &offset) {
        Position = Position + offset;
    }

    void Sphere::Transform(const Matrix3x3 &transform) {
        Position = transform * Position;
    }

    void Sphere::Transform(const Matrix4x4 &transform) {
        Position = transform * Position;
    }

    float Sphere::Cube(float inp) {
        return inp*inp*inp;
    }

    float Sphere::Volume() const {
        return 4.f * Math::Pi * Cube(Radius) / 3.f;
    }

    float Sphere::SurfaceArea() const {
        return 4.f * Math::Pi * Cube(Radius) / 3.f;
    }

    bool Sphere::IsFinite() const {
        return Position.IsFinite() && std::isfinite(Radius);
    }

    bool Sphere::IsDegenerate() const {
        return !(Radius > 0.f) || !Position.IsFinite();
    }

    bool Sphere::Contains(const Vector3 &point) const {
        return Position.DistanceSquared(point) <= Radius*Radius;
    }

    bool Sphere::Contains(const Vector3 &point, float epsilon) const {
        return Position.DistanceSquared(point) <= Radius*Radius + epsilon;
    }

    Vector3 Sphere::RandomPointOnSurface(RNG &rng) const {
        Vector3 v = Vector3::Zero;
        // Rejection sampling analysis: The unit sphere fills ~52.4% of the volume of its enclosing box,
        // so this loop is expected to take only very few iterations before succeeding.
        for (int i = 0; i < 1000; ++i)
        {
            v.x = rng.FloatNeg1_1();
            v.y = rng.FloatNeg1_1();
            v.z = rng.FloatNeg1_1();
            float lenSq = v.LengthSquared();
            if (lenSq >= 1e-6f && lenSq <= 1.f)
                return Position + (Radius / std::sqrt(lenSq)) * v;
        }
        // Astronomically small probability to reach here, and if we do so, the provided RNG must have been in a bad state.
        assert(false && "Sphere::RandomPointOnSurface failed!");

        // Failed to generate a point inside this sphere. Return an arbitrary point on the surface as fallback.
        return Position + Vector3(Radius, 0, 0);
    }

    Vector3 Sphere::RandomPointInside(RNG &rng) {
        assert(Radius > 1e-3f);
        Vector3 v = Vector3::Zero;

        // Rejection sampling analysis: The unit sphere fills ~52.4% of the volume of its enclosing box
        // so this loop is expected to take only very few iterations before succeeding.
        for (int i = 0; i < 1000; ++i)
        {
            v.x = rng.Float(-Radius, Radius);
            v.y = rng.Float(-Radius, Radius);
            v.z = rng.Float(-Radius, Radius);

            if (v.LengthSquared() <= Radius*Radius)
                return Position + v;
        }
        assert(false && "Sphere::RandomPointInside failed!");

        // Failed to generate a point inside this sphere. Return the sphere center as fallback.
        return Position;
    }


}
