#include <J3ML/LinearAlgebra/Vector2.hpp>
#include <cassert>
#include <algorithm>
#include <valarray>
#include <iostream>
#include <format>
#include <J3ML/LinearAlgebra/Matrix3x3.hpp>
#include <J3ML/LinearAlgebra/Matrix4x4.hpp>
#include <J3ML/LinearAlgebra/Vector4.hpp>

namespace J3ML::LinearAlgebra {

    Vector2::Vector2()
    {}

    Vector2::Vector2(float X, float Y): x(X), y(Y)
    {}

    Vector2::Vector2(const Vector2& rhs): x(rhs.x), y(rhs.y)
    {}

    float Vector2::operator[](std::size_t index) const
    {
        return At(index);
    }

    float &Vector2::operator[](std::size_t index)
    {
        return At(index);
    }

    bool Vector2::IsWithinMarginOfError(const Vector2& rhs, float margin) const
    {
        return this->Distance(rhs) <= margin;
    }

    bool Vector2::operator==(const Vector2& rhs) const
    {
        return this->IsWithinMarginOfError(rhs);
    }

    bool Vector2::operator!=(const Vector2& rhs) const
    {
        return this->IsWithinMarginOfError(rhs) == false;
    }

    Vector2 Vector2::Min(const Vector2& min) const
    {
        return {
                std::min(this->x, min.x),
                std::min(this->y, min.y)
        };
    }

    Vector2 Vector2::Max(const Vector2& max) const
    {
        return {
                std::max(this->x, max.x),
                std::max(this->y, max.y)
        };
    }

    Vector2 Vector2::Clamp(const Vector2& min, const Vector2& max) const
    {
        return {
                std::clamp(this->x, min.x, max.x),
                std::clamp(this->y, min.y, max.y)
        };
    }

    float Vector2::Distance(const Vector2& to) const
    {
        return ((*this)-to).Magnitude();
    }

    float Vector2::Length() const
    {
        return std::sqrt(LengthSquared());
    }

    float Vector2::LengthSquared() const
    {
        return (x*x + y*y);
    }

    float Vector2::Magnitude() const
    {
        return std::sqrt(LengthSquared());
    }

    float Vector2::Dot(const Vector2& rhs) const
    {
        return this->x * rhs.x + this->y * rhs.y;
    }

    Vector2 Vector2::Project(const Vector2& rhs) const
    {
        float scalar = this->Dot(rhs) / (rhs.Magnitude()*rhs.Magnitude());
        return rhs * scalar;
    }

    Vector2 Vector2::Normalized() const
    {
        Vector2 copy = *this;
        float oldLength = copy.Normalize();
        assert(oldLength > 0.f && "Vector2::Normalized() failed!");
        return copy;
    }

    Vector2 Vector2::Lerp(const Vector2& rhs, float alpha) const
    {
        return this->operator*(1.0f - alpha) + (rhs * alpha);
    }

    float Vector2::AngleBetween(const Vector2& rhs) const
    {
        auto numer = this->Dot(rhs);
        auto denom = this->Magnitude() * rhs.Magnitude();
        return std::acos(numer / denom);
    }

    float Vector2::AngleBetween(const Vector2& lhs, const Vector2& rhs)
    { return lhs.AngleBetween(rhs); }

    Vector2 Vector2::operator+(const Vector2& rhs) const
    {
        return {this->x + rhs.x, this->y + rhs.y};
    }

    Vector2 Vector2::operator-(const Vector2& rhs) const
    {
        return {this->x - rhs.x, this->y - rhs.y};
    }

    Vector2 Vector2::operator*(float rhs) const
    {
        return {
                this->x * rhs,
                this->y * rhs
        };
    }

    Vector2 Vector2::operator/(float rhs) const
    {
        return {
                this->x / rhs,
                this->y / rhs
        };
    }

    Vector2 Vector2::operator-() const
    {
        return {-x, -y};
    }

    const Vector2 Vector2::Zero  = {0, 0};
    const Vector2 Vector2::One   = {1, 1};
    const Vector2 Vector2::UnitX = {1, 0};
    const Vector2 Vector2::UnitY = {0, 1};
    const Vector2 Vector2::Up    = {0, -1};
    const Vector2 Vector2::Down  = {0, 1};
    const Vector2 Vector2::Left  = {-1, 0};
    const Vector2 Vector2::Right = {1, 0};
    const Vector2 Vector2::NaN = {NAN, NAN};
    const Vector2 Vector2::Infinity = {INFINITY, INFINITY};

    float Vector2::GetX() const { return x; }

    float Vector2::GetY() const { return y; }

    float Vector2::LengthSquared(const Vector2 &of) { return of.LengthSquared(); }

    Vector2 Vector2::Min(const Vector2 &value, const Vector2 &minimum) { return value.Min(minimum); }

    Vector2 Vector2::Max(const Vector2 &value, const Vector2 &maximum) { return value.Max(maximum);}

    float Vector2::Length(const Vector2 &of) { return of.Length(); }

    void Vector2::SetX(float newX) { x = newX;}

    void Vector2::SetY(float newY) { y = newY; }

    bool Vector2::IsNormalized(float epsilonSq) const {
        return std::abs(LengthSquared() - 1.f) <= epsilonSq;
    }

    bool Vector2::IsZero(float epsilonSq) const {
        return LengthSquared() <= epsilonSq;
    }

    bool Vector2::IsPerpendicular(const Vector2 &other, float epsilonSq) const {
        float dot = Dot(other);
        return dot*dot <= epsilonSq * LengthSquared() * other.LengthSquared();
    }



    Vector2 Vector2::Project(const Vector2 &lhs, const Vector2 &rhs) { return lhs.Project(rhs); }

    float Vector2::Dot(const Vector2 &lhs, const Vector2 &rhs) { return lhs.Dot(rhs); }

    float Vector2::Magnitude(const Vector2 &of) { return of.Magnitude();}

    Vector2 Vector2::Lerp(const Vector2 &lhs, const Vector2 &rhs, float alpha) { return lhs.Lerp(rhs, alpha); }

    Vector2 Vector2::Div(const Vector2 &lhs, float rhs) {
        return lhs.Div(rhs);
    }

    Vector2 Vector2::Mul(const Vector2 &lhs, float rhs) {
        return lhs.Mul(rhs);
    }

    Vector2 Vector2::Sub(const Vector2 &lhs, const Vector2 &rhs) {
        return lhs.Sub(rhs);
    }

    Vector2 Vector2::Add(const Vector2 &lhs, const Vector2 &rhs) {
        return lhs.Add(rhs);
    }

    Vector2 Vector2::Add(const Vector2 &rhs) const {
        return *this + rhs;
    }

    Vector2 Vector2::Sub(const Vector2 &rhs) const {
        return *this - rhs;
    }

    Vector2 Vector2::Mul(float scalar) const {
        return *this * scalar;
    }

    Vector2 Vector2::Div(float scalar) const {
        return *this / scalar;
    }

    bool Vector2::IsFinite(const Vector2 &v) {
        return v.IsFinite();
    }

    float Vector2::MinElement() const {
        return std::min(x, y);
    }

    float Vector2::MaxElement() const {
        return std::max(x, y);
    }

    Vector2 Vector2::Mul(const Vector2 &v) const {
        return {this->x*v.x, this->y*v.y};
    }

    bool Vector2::IsFinite() const {
        return std::isfinite(x) && std::isfinite(y);
    }

    Vector2 Vector2::Div(const Vector2 &v) const {
        return {this->x/v.x, this->y/v.y};
    }

    Vector2 Vector2::Abs() const { return {std::abs(x), std::abs(y)};}

    float *Vector2::ptr() {
        return &x;
    }

    const float *Vector2::ptr() const { return &x;}

    const float Vector2::At(std::size_t index) const {
        assert(index >= 0);
        assert(index < Dimensions);
        return ptr()[index];
    }

    float &Vector2::At(std::size_t index) {
        assert(index >= 0);
        assert(index < Dimensions);
        return ptr()[index];
    }

    Vector2 &Vector2::operator/=(float scalar) {
        x /= scalar;
        y /= scalar;

        return *this;
    }

    Vector2 &Vector2::operator*=(float scalar) {
        x *= scalar;
        y *= scalar;

        return *this;
    }

    Vector2 &Vector2::operator-=(const Vector2 &rhs) // Subtracts a vector from this vector, in-place
    {
        x -= rhs.x;
        y -= rhs.y;

        return *this;
    }

    Vector2 &Vector2::operator+=(const Vector2 &rhs) // Adds a vector to this vector, in-place.
    {
        x += rhs.x;
        y += rhs.y;

        return *this;
    }

    Vector2 &Vector2::operator=(const Vector2 &rhs) {
        x = rhs.x;
        y = rhs.y;

        return *this;
    }

    Vector2::Vector2(const float *data) {
        assert(data);
        x = data[0];
        y = data[1];
    }

    Vector2::Vector2(float scalar) {
        x = scalar;
        y = scalar;
    }

    float Vector2::DistanceSq(const Vector2 &to) const {
        return (*this-to).LengthSquared();
    }

    float Vector2::DistanceSq(const Vector2 &from, const Vector2 &to) {
        return (from-to).LengthSquared();
    }

    bool Vector2::OrientedCCW(const Vector2 &a, const Vector2 &b, const Vector2 &c)
    {
        // Compute the determinant
        // | ax ay 1 |
        // | bx by 1 |
        // | cx cy 1 |
        // See Christer Ericson, Real-Time Collision Detection, p.32.
        return (a.x-c.x)*(b.y-c.y) - (a.y-c.y)*(b.x-c.x) >= 0.f;
    }

    Vector2 Vector2::operator*(const Vector2 &rhs) const {
        return this->Mul(rhs);
    }

    Vector2 Vector2::operator/(const Vector2 &rhs) const { return this->Div(rhs);}

    Vector2 Vector2::Clamp(float floor, float ceil) const {
        return {std::clamp(this->x, floor, ceil),
                       std::clamp(this->y, floor, ceil)};
    }

    Vector2 Vector2::Clamp(const Vector2 &value, float floor, float ceil) {
        return value.Clamp(floor, ceil);
    }

    Vector2 Vector2::Clamp01() const {
        return Clamp(0, 1);
    }

    Vector2 Vector2::Clamp01(const Vector2 &value) {
        return value.Clamp01();
    }

    Vector2 Vector2::Clamp(const Vector2 &min, const Vector2 &middle, const Vector2 &max) {
        return middle.Clamp(min, max);
    }

    float Vector2::Distance(const Vector2 &from, const Vector2 &to) {
        return from.Distance(to);
    }

    float Vector2::SumOfElements() const { return x+y;}

    float Vector2::ProductOfElements() const { return x*y;}

    float Vector2::AverageOfElements() const { return (x+y)/2; }

    Vector2 Vector2::Neg() const { return {-x, -y};}

    Vector2 Vector2::Recip() const { return {1.f/x, 1.f/y}; }

    float Vector2::AimedAngle() const {
        assert(!IsZero());
        return std::atan2(y, x);
    }

    Vector2 Vector2::ToPolarCoordinates() const {
        float radius = Length();
        if (radius > 1e-4f)
            return {std::atan2(y, x), radius};
        else
            return Vector2::Zero;
    }

    Vector2 Vector2::Perp() const {
        return {-y, x};
    }

    float Vector2::PerpDot(const Vector2 &rhs) const {
        return x * rhs.y - y * rhs.x;
    }

    void Vector2::Rotate90CW() {
        float oldX = x;
        x = y;
        y = -oldX;
    }

    Vector2 Vector2::Rotated90CW() const {
        return {y, -x};
    }

    void Vector2::Rotate90CCW() {
        float oldX = x;
        x = -y;
        y = oldX;
    }

    Vector2 Vector2::Rotated90CCW() const {
        return {-y, x};
    }

    Vector2 Vector2::Reflect(const Vector2 &normal) const {
        assert(normal.IsNormalized());
        return 2.f * this->ProjectToNorm(normal) - *this;
    }

    Vector2
    Vector2::Refract(const Vector2 &normal, float negativeSideRefractionIndex, float positiveSideRefractionIndex) const {
        // Implementation from https://www.flipcode.com/archives/reflection_transmission.pdf.
        // This code is duplicated in Vector3::Refract
        float n = negativeSideRefractionIndex / positiveSideRefractionIndex;
        float cosI = this->Dot(normal);
        float sinT2 = n*n*(1.f - cosI*cosI);
        if (sinT2 > 1.f) // Total internal reflection occurs?
            return (-*this).Reflect(normal);
        return n * *this - (n + std::sqrt(1.f - sinT2)) * normal;
    }

    Vector2 Vector2::ProjectTo(const Vector2 &direction) const {
        assert(!direction.IsZero());
        return direction * this->Dot(direction) / direction.LengthSquared();
    }

    Vector2 Vector2::ProjectToNorm(const Vector2 &direction) const {
        assert(direction.IsNormalized());
        return direction * this->Dot(direction);
    }

    void Vector2::Orthogonalize(const Vector2 &a,  Vector2 &b) {
        assert(!a.IsZero());
        b -= a.Dot(b) / a.Length() * a;
    }

    void Vector2::Orthonormalize(Vector2 &a, Vector2 &b) {
        assert(!a.IsZero());
        a.Normalize();
        b -= a.Dot(b) * a;
    }

    Vector2 Vector2::RandomBox(Algorithm::RNG &rng, float minElem, float maxElem) {
        float x = rng.Float(minElem, maxElem);
        float y = rng.Float(minElem, maxElem);

        return {x, y};
    }

    bool Vector2::AreOrthogonal(const Vector2 &a, const Vector2 &b, float epsilon) {
        return a.IsPerpendicular(b, epsilon);
    }

    void Vector2::Set(float newX, float newY) {
        x = newX;
        y = newY;
    }

    float Vector2::Normalize() {
        assert(IsFinite());
        float lengthSq = LengthSquared();
        if (lengthSq > 1e-6f)
        {
            float length = std::sqrt(lengthSq);
            *this *= 1.f / length;
            return length;
        } else {
            Set(1.f, 0.f); // We will always produce a normalized vector.
            return 0; // But signal failure, so user knows we have generated an arbitrary normalization.
        }
    }

    std::string Vector2::ToString() const {
        return std::format("{},{}", x, y);
    }

    Vector2 operator*(float lhs, const Vector2 &rhs) {
        return {lhs * rhs.x, lhs * rhs.y};
    }

    Vector2 Mul2D(const Matrix3x3 &transform, const Vector2 &v) { return transform.Transform(v);}

    std::ostream &operator<<(std::ostream &out, const Vector2 &rhs) {
        std::string str = rhs.ToString();
        out << str;
        return out;
    }

    Vector2 MulPos2D(const Matrix4x4 &transform, const Vector2 &v) {
        return transform.Transform(Vector4(v.x, v.y, 0.f, 1.f)).XY();
    }

    Vector2 MulDir2D(const Matrix4x4 &transform, const Vector2 &v) {
        return transform.Transform(Vector4(v.x, v.y, 0.f, 0.f)).XY();
    }
}