
#pragma region vector4


#include <J3ML/LinearAlgebra/Vector4.hpp>
#include <J3ML/LinearAlgebra/Vector3.hpp>
#include <cmath>
#include <algorithm>

namespace J3ML::LinearAlgebra {

    using namespace J3ML::Math;


    const Vector4 Vector4::Zero = {0,0,0,0};
    const Vector4 Vector4::NaN = {NAN, NAN, NAN, NAN};

    Vector4::Vector4(): x(0), y(0), z(0), w(0)
    {}

    Vector4::Vector4(float X, float Y, float Z, float W): x(X),y(Y),z(Z),w(W)
    { }

    bool Vector4::IsFinite() const {
        return std::isfinite(x) && std::isfinite(y) && std::isfinite(z) && std::isfinite(w);
    }

    bool Vector4::IsPerpendicular(const Vector4 &other, float epsilonSq) const {
        float dot = Dot(other);
        return dot*dot <= epsilonSq * LengthSquared() * other.LengthSquared();
    }

    bool Vector4::IsPerpendicular3(const Vector4 &other, float epsilonSq) const {
        float dot = Dot3(other);
        return dot*dot <= epsilonSq * LengthSq() * other.LengthSq();
    }

    void Vector4::NormalizeW() {
        if (std::abs(w) > 1e-6f) {
            float invW = 1.f / w;
            x *= invW;
            y *= invW;
            z *= invW;
            w = 1.f;
        }
    }

bool Vector4::operator==(const Vector4& rhs) const
{
    return this->IsWithinMarginOfError(rhs);
}

bool Vector4::operator!=(const Vector4& rhs) const
{
    return this->IsWithinMarginOfError(rhs) == false;
}

Vector4 Vector4::Min(const Vector4& min) const
{
    return {
            std::min(this->x, min.x),
            std::min(this->y, min.y),
            std::min(this->z, min.z),
            std::min(this->w, min.w)
    };
}

Vector4 Vector4::Max(float floor) const {
    return {std::max(x, floor), std::max(y, floor), std::max(z, floor), std::max(w, floor)};
}

Vector4 Vector4::Max(const Vector4& max) const
{
    return {
            std::max(this->x, max.x),
            std::max(this->y, max.y),
            std::max(this->z, max.z),
            std::max(this->w, max.w)
    };
}

Vector4 Vector4::Clamp(float floor, float ceil) const {
    return {Math::Clamp(x, floor, ceil),
            Math::Clamp(y, floor, ceil),
            Math::Clamp(z, floor, ceil),
            Math::Clamp(w, floor, ceil)
    };
}

Vector4 Vector4::Clamp(const Vector4& min, const Vector4& max) const
{
    return {
            std::clamp(this->x, min.x, max.x),
            std::clamp(this->y, min.y, max.y),
            std::clamp(this->z, min.z, max.z),
            std::clamp(this->w, min.w, max.w)
    };
}

float Vector4::Distance(const Vector4& to) const
{
    return ( (*this) - to ).Magnitude();
}

float Vector4::Length4() const {
    return std::sqrt(x*x + y*y + z*z + w*w);
}

float Vector4::Length() const { return Length4(); }

float Vector4::LengthSquared4() const {
    return x*x + y*y + z*z + w*w;
}

float Vector4::LengthSquared() const
{
    return (x*x + y*y + z*z + w*w);
}

float Vector4::LengthSq4() const { return LengthSquared4();}

float Vector4::LengthSq() const { return LengthSquared();}

float Vector4::LengthSquared3() const { return LengthSq3(); }

float Vector4::LengthSq3() const { return x*x + y*y + z*z; }

float Vector4::Magnitude() const
{
    return std::sqrt(x*x + y*y + z*z + w*w);
}

float Vector4::Dot(const Vector4& rhs) const
{
    auto a = this->Normalized();
    auto b = rhs.Normalized();

    return  a.x * b.x +
            a.y * b.y +
            a.z * b.z +
            a.w * b.w;
}

float Vector4::Dot3(const Vector3 &rhs) const {
    return x * rhs.x + y * rhs.y + z * rhs.z;
}

float Vector4::Dot3(const Vector4 &rhs) const {
    return x * rhs.x + y * rhs.y + z * rhs.z;
}

Vector4 Vector4::Project(const Vector4& rhs) const
{
    float scalar = this->Dot(rhs) / (rhs.Magnitude()* rhs.Magnitude());
    return rhs * scalar;
}

void Vector4::Normalize()
{
    if (Length() > 0) {
        x = x / Length();
        y = y / Length();
        z = z / Length();
        w = w / Length();
    }
}

Vector4 Vector4::Normalized() const
{
    if (Length() > 0)
        return {
                x / Length(),
                y / Length(),
                z / Length(),
                w / Length()
        };
    else
        return {0,0,0,0};
}

Vector4 Vector4::Lerp(const Vector4& goal, float alpha) const
{
    return this->operator*(1.0f - alpha) + (goal * alpha);
}

Vector4 Vector4::operator+(const Vector4& rhs) const
{
    return {x+rhs.x, y+rhs.y, z+rhs.z, w+rhs.w};
}

Vector4 Vector4::operator-(const Vector4& rhs) const
{
    return {x-rhs.x, y-rhs.y, z-rhs.z, w-rhs.w};
}

    Vector4 Vector4::operator*(float rhs) const {
        return {
                this->x * rhs,
                this->y * rhs,
                this->z * rhs,
                this->w * rhs
        };
    }

    bool Vector4::IsWithinMarginOfError(const Vector4 &rhs, float margin) const {
        return this->Distance(rhs) <= margin;
    }

    Vector4::Vector4(const Vector3 &xyz, float w) : x(xyz.x), y(xyz.y), z(xyz.z), w(w)
    {

    }

    Vector4 &Vector4::operator=(const Vector4 &rhs) {
        x = rhs.x;
        y = rhs.y;
        z = rhs.z;
        w = rhs.w;

        return *this;
    }

    Vector4::Vector4(const float *data) {
        assert(data);
        x = data[0];
        y = data[1];
        z = data[2];
        w = data[3];
    }

    float * Vector4::ptr() { return &x; }

    const float * Vector4::ptr() const { return &x; }

    Vector2 Vector4::XY() const { return {x,y}; }

    Vector3 Vector4::XYZ() const { return {x, y, z}; }

    void Vector4::Set(float x_, float y_, float z_, float w_) {
        this->x = x_;
        this->y = y_;
        this->z = z_;
        this->w = w_;
    }

    void Vector4::Set(const Vector4 &rhs) {
        x = rhs.x;
        y = rhs.y;
        z = rhs.z;
        w = rhs.w;
    }

    bool Vector4::Equals(const Vector4 &rhs, float epsilon) const {
        return std::abs(x - rhs.x) < epsilon &&
               std::abs(y - rhs.y) < epsilon &&
               std::abs(z - rhs.z) < epsilon &&
               std::abs(w - rhs.w) < epsilon;
    }

    bool Vector4::Equals(float _x, float _y, float _z, float _w, float epsilon) const {
        return std::abs(x - _x) < epsilon &&
               std::abs(y - _y) < epsilon &&
               std::abs(z - _z) < epsilon &&
               std::abs(w - _w) < epsilon;
    }

    Vector4 Vector4::Min(float ceil) const {
        return {std::min(x, ceil), std::min(y, ceil), std::min(z, ceil), std::min(w, ceil)};
    }

    float Vector4::operator[](std::size_t index) const {
        assert(index < 4);
        if (index==0) return x;
        if (index==1) return y;
        if (index==2) return z;
        if (index==3) return w;
        return 0;
    }

    float &Vector4::operator[](std::size_t index) {
        assert(index < 4);
        if (index == 0) return x;
        if (index == 1) return y;
        if (index == 2) return z;
        if (index == 3) return w;

        return x; // This point should never be reached.
    }

    float Vector4::At(int index) const {
        assert(index >= 0);
        assert(index < Size);
        return ptr()[index];
    }

    float & Vector4::At(int index) {
        assert(index >= 0);
        assert(index < Size);
        return ptr()[index];
    }

    float Vector4::LengthSqXYZ() const {
        return x*x * y*y * z*z;
    }

    bool Vector4::IsNormalized3(float epsilonSq) const {
        return std::abs(LengthSqXYZ()-1.f) <= epsilonSq;
    }

    bool Vector4::IsNormalized4(float epsilonSq) const {
        return std::abs(LengthSquared()-1.f) <= epsilonSq;
    }

    bool Vector4::IsNormalized(float epsilonSq) const { return IsNormalized4(epsilonSq); }

    bool Vector4::IsZero3(float epsilonSq) const {
        return LengthSq3() <= epsilonSq;
    }

    Vector4 Vector4::Cross3(const Vector3 &rhs) const {
        Vector4 dst;
        dst.x = y * rhs.z - z * rhs.y;
        dst.y = z * rhs.x - x * rhs.z;
        dst.z = x * rhs.y - y * rhs.x;
        dst.w = 0.f;
        return dst;
    }

    Vector4 Vector4::Cross3(const Vector4 &rhs) const {
        return Cross3(rhs.XYZ());
    }

    Vector4 Vector4::Cross(const Vector4 &rhs) const { return Cross3(rhs); }
}
#pragma endregion